#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <cstring>

#include "filename/filename.h"
#include "opt/opt.h"
#include "err/err.h"
#include "time_fmt/time_fmt.h"
#include <libxml/xmlreader.h>
#include <libxml/xmlwriter.h>

#include "geo_data.h"

using namespace std;

/**
 libxml2 manual: http://www.xmlsoft.org/html/index.html
 xmlreader module: http://www.xmlsoft.org/html/libxml-xmlreader.html
 xmlwriter module: http://www.xmlsoft.org/html/libxml-xmlwriter.html
 GPX 1.1 format: http://www.topografix.com/GPX/1/1/
 GPX 1.0 format: http://www.topografix.com/gpx_manual.asp
 example with all fields: https://github.com/tkrajina/gpxgo/tree/master/test_files

GPX 1.1 Structure:
<gpx version creator>
  <metadata>
    <name> <desc> <author> <copyright> <link> <time> <keywords> <bounds> <extensions>
  <wpt lat lon>
    <ele> <time> <magvar> <geoidheight>
    <name> <cmt> <desc> <src> <link> <sym> <type>
    <fix> <sat> <hdop> <vdop> <pdop> <ageofdgpsdata>
    <dgpsid> <extensions>

  <rte>
    <name> <cmt> <desc> <src> <link> <number> <type> <extensions>
    <rtept> [0..*] -- same as <wpt>

  <trk>
    <name> <cmt> <desc> <src> <link> <number> <type> <extensions> -- same as in rte
    <trkseg> [0..*]
      <trkpt> [0..*] -- same as <wpt>
      <extensions>
  <extensions>

GPX 1.0 Structure:
<gpx version creator>
  <name> <desc> <author> <email> <url> <urlname> <time> <keywords> <bounds>
  <wpt lat lon>
    <ele> <time> <magvar> <geoidheight>
    <name> <cmt> <desc> <src> <url> <urlname> <sym> <type>
    <fix> <sat> <hdop> <vdop> <pdop> <ageofdgpsdata>
    <dgpsid>

  <rte>
    <name> <cmt> <desc> <src> <url> <urlname> <number>
    <rtept> [0..*] -- same as <wpt>

  <trk>
    <name> <cmt> <desc> <src> <url> <urlname> <number> -- same as in rte
    <trkseg> [0..*]
      <trkpt> [0..*] -- same as <wpt>


-----------------
<wpt> fields:

Attributes: lat, lon -- The latitude and longitude of the point. Decimal
degrees, WGS84 datum. (x and y fields in the mapsoft GeoWpt structure).
Two mandatory fields.

* ele -- Elevation (in meters) of the point. (z field in the mapsoft
  GeoWpt structure).
* time -- Creation/modification timestamp for element. Date and time in
  are in Univeral Coordinated Time (UTC), not local time! Conforms to ISO
  8601 specification for date/time representation. Fractional seconds are
  allowed for millisecond timing in tracklogs. (opt["time"] in the mapsoft
  GeoWpt structure).

All other elements are stored in the Opt object and handeled as strings.

* magvar -- Magnetic variation (in degrees) at the point.
* geoidheight -- Height (in meters) of geoid (mean sea level) above WGS84
  earth ellipsoid. As defined in NMEA GGA message.
* name -- The GPS name of the waypoint. This field will be transferred to and from
  the GPS. GPX does not place restrictions on the length of this field or
  the characters contained in it. It is up to the receiving application to
  validate the field before sending it to the GPS.
* cmt -- GPS waypoint comment. Sent to GPS as comment. In mapsoft2 field name is "comm".
* desc -- A text description of the element. Holds additional information
  about the element intended for the user, not the GPS.
* src -- Source of data. Included to give user some idea of reliability
  and accuracy of data. "Garmin eTrex", "USGS quad Boston North", e.g.
* link -- Link to additional information about the waypoint.
* sym -- Text of GPS symbol name. For interchange with other programs,
  use the exact spelling of the symbol as displayed on the GPS. If the GPS
  abbreviates words, spell them out.
* type -- Type (classification) of the waypoint.
* fix -- Type of GPX fix. (none, 2d, 3d, dgps, pps)
* sat -- Number of satellites used to calculate the GPX fix.
* hdop -- Horizontal dilution of precision.
* vdop -- Vertical dilution of precision.
* pdop -- Position dilution of precision.
* ageofdgpsdata -- Number of seconds since last DGPS update.
* dgpsid -- ID of DGPS station used in differential correction.
* extensions -- You can add extend GPX by adding your own elements from
  another schema here.

    Rte/Trk attributes:

* name -- GPS name of route.
* cmt -- GPS comment for route. In mapsoft2 field name is "comm".
* desc -- Text description of route for user. Not sent to GPS.
* src -- Source of data. Included to give user some idea of reliability
  and accuracy of data.
* link -- Links to external information about the route.
* number -- GPS route number.
* type -- Type (classification) of route.
* extensions -- You can add extend GPX by adding your own elements from
  another schema here.

Element order is important, see:
https://github.com/wladich/nakarte/issues/386

-----------------
    support in Mapsoft:

reading:
- skip gpx attributes, <metadata> and all <extensions>
- rte -- load as waypoint list, everything except extensions (TODO: test!)
- wpt: everything except extensions, waypoint list name is set to
       filename without .gpx extension
- trk: everything except extensions
- - trkseg: everything except extensions
- - - trkpt: only <lat> <lon> <ele> <time> tags

writing:
- skip gpx attributes, <metadata> and all <extensions>
- wpt - all waypoint lists are written together.
        everithing except extensions
- trk - everithing except extensions
- - trkseg: everything except extensions
- - - trkpt: only <lat> <lon> <ele> <time> tags
*/


// All names which are keps in opt, handeled as strings
// and have same spelling in GPX and in mapsoft.
// comm/cmt, time, ele/z are processed separately.
const char *gps_wpt_names[] = {
  "magvar", "geoidheight", "desc", "src", "link", "sym", "type", "fix",
  "sat", "hdop", "vdop", "pdop", "ageofdgpsdata", "dgpsid", NULL};

/// Trk/rte fields.
/// comm/cmt processed separately.
const char *gps_trk_names[] = {
  "desc", "src", "link",
  "number", "type", NULL};

/********************************************************************/
/// Write GPX.
/// options:
///   xml_compr:   compress the output? 0|1, default 0;
///   xml_indent:  use indentation? 0|1, default 1;
///   xml_ind_str: indentation string, default "  ";
///   xml_qchar:   quoting character for attributes, default \'
///   gpx_write_rte: write waypoint lists as routes, 0|1, default 0
void
write_gpx (const string &filename, const GeoData & data, const Opt & opts){

  LIBXML_TEST_VERSION

  // set some parameters
  int indent = opts.get<int>("xml_indent", 1);
  char qchar = opts.get<char>("xml_qchar", '\'');
  std::string ind_str = opts.get("xml_ind_str", "  ");
  bool v = opts.get("verbose", false);
  if (v) cerr << "Writing GPX file: " << filename << endl;

  // create XML writer
  xmlTextWriterPtr writer =
    xmlNewTextWriterFilename(filename.c_str(), opts.get<int>("xml_compr", 0));
  if (writer == NULL)
    throw Err() << "write_gpx: can't write to file: " << filename;

  try {
    if (xmlTextWriterSetIndent(writer, indent)<0 ||
        xmlTextWriterSetIndentString(writer, BAD_CAST ind_str.c_str())<0 ||
        xmlTextWriterSetQuoteChar(writer, qchar)<0)
      throw "setting xml writer parameters";

    // start XML document
    if (xmlTextWriterStartDocument(writer, "1.0", "UTF-8", NULL)<0)
      throw "starting the xml document";

    // start GPX element.
    // BAD_CAST converts (const char*) to xmlChar*
    if (xmlTextWriterStartElement(writer, BAD_CAST "gpx")<0 ||
        xmlTextWriterWriteAttribute(writer,
          BAD_CAST "version", BAD_CAST "1.1")<0 ||
        xmlTextWriterWriteAttribute(writer,
          BAD_CAST "creator", BAD_CAST "mapsoft-2")<0)
      throw "starting <gpx> element";

    int use_rte = opts.get<int>("gpx_write_rte", 0);

    // Writing waypoints:
    for (auto wpl: data.wpts) {

      if (v) cerr << "  Writing " << (use_rte? "waypoints":"route") << ": "
                  << wpl.name << " (" << wpl.size() << " points)" << endl;

      if (use_rte){
        if (xmlTextWriterStartElement(writer, BAD_CAST "rte")<0)
          throw "starting <rte> element";

        if (wpl.name!="" &&
            xmlTextWriterWriteFormatElement(writer,
            BAD_CAST "name", "%s", wpl.name.c_str())<0)
              throw "writing <name> element";

        // other option elements:
        for (const char **fn = gps_trk_names; *fn!=NULL; fn++){
          string pfn = string("gpx_") + (*fn);
          if (!wpl.opts.exists(pfn)) continue;
          if (xmlTextWriterWriteFormatElement(writer,
             BAD_CAST *fn, "%s", wpl.opts.get<string>(pfn).c_str())<0)
            throw "writing element";
        }
      }
      for (auto wp: wpl) {

        const xmlChar* pt = BAD_CAST (use_rte? "rtept" : "wpt");
        if (xmlTextWriterStartElement(writer, pt)<0 ||
            xmlTextWriterWriteFormatAttribute(writer,
              BAD_CAST "lat", "%.7f", wp.y)<0 ||
            xmlTextWriterWriteFormatAttribute(writer,
              BAD_CAST "lon", "%.7f", wp.x)<0)
          throw "starting <wpt> element";

        // altitude
        if (wp.have_alt() &&
            xmlTextWriterWriteFormatElement(writer,
              BAD_CAST "ele", "%.2f", wp.z)<0)
          throw "writing <ele> element";

        // time
        if (wp.t > 0 &&
            xmlTextWriterWriteFormatElement(writer,
              BAD_CAST "time", "%s", write_fmt_time("%FT%T%fZ", wp.t).c_str())<0)
          throw "writing <time> element";

        // name
        if (wp.name != "" &&
            xmlTextWriterWriteFormatElement(writer,
              BAD_CAST "name", "%s", wp.name.c_str())<0)
          throw "writing <name> element";

        // cmt
        if (wp.comm != "" &&
            xmlTextWriterWriteFormatElement(writer,
              BAD_CAST "cmt", "%s", wp.comm.c_str())<0)
          throw "writing <cmt> element";

        // other option elements:
        for (const char **fn = gps_wpt_names; *fn!=NULL; fn++){
          string pfn = string("gpx_") + (*fn);
          if (!wp.opts.exists(pfn)) continue;
          if (xmlTextWriterWriteFormatElement(writer,
             BAD_CAST *fn, "%s", wp.opts.get<string>(pfn).c_str())<0)
            throw "writing element";
        }

        if (xmlTextWriterEndElement(writer) < 0)
          throw "closing <wpt> element";
      }
      if (use_rte){
        if (xmlTextWriterEndElement(writer)<0)
          throw "closing <rte> element";
      }
    }


    // Write tracks:
    for (auto trk: data.trks) {

      if (v) cerr << "  Writing track: "
                  << trk.name << " (" << trk.size() << " points)" << endl;

      if (xmlTextWriterStartElement(writer, BAD_CAST "trk")<0)
          throw "starting <trk> element";

      // name
      if (trk.name != "" &&
        xmlTextWriterWriteFormatElement(writer,
           BAD_CAST "name", "%s", trk.name.c_str())<0)
        throw "writing <name> element";

      // other option elements:
      for (const char **fn = gps_trk_names; *fn!=NULL; fn++){
        string pfn = string("gpx_") + (*fn);
        if (!trk.opts.exists(pfn)) continue;
        if (xmlTextWriterWriteFormatElement(writer,
           BAD_CAST *fn, "%s", trk.opts.get<string>(pfn).c_str())<0)
          throw "writing element";
      }

      int cnt=0;
      for (auto tp: trk) {
        cnt++;

        // close <trkseg> if needed
        if (tp.start && cnt>1 &&
           xmlTextWriterEndElement(writer) < 0)
             throw "closing <trkseg> element";

        // open <trkseg> if needed
        if ((tp.start || cnt==1) &&
           xmlTextWriterStartElement(writer, BAD_CAST "trkseg") < 0)
             throw "starting <trkseg> element";

        // open <trkpt>
        if (xmlTextWriterStartElement(writer, BAD_CAST "trkpt")<0 ||
            xmlTextWriterWriteFormatAttribute(writer,
              BAD_CAST "lat", "%.7f", tp.y)<0 ||
            xmlTextWriterWriteFormatAttribute(writer,
              BAD_CAST "lon", "%.7f", tp.x)<0)
          throw "starting <trkpt> element";

        // altitude
        if (tp.have_alt() &&
            xmlTextWriterWriteFormatElement(writer,
              BAD_CAST "ele", "%.2f", tp.z)<0)
          throw "writing <ele> element";

        // time
        if (tp.t && xmlTextWriterWriteFormatElement(writer,
              BAD_CAST "time", "%s",
                 write_fmt_time("%FT%T%fZ", tp.t).c_str())<0)
          throw "writing <time> element";


        if (xmlTextWriterEndElement(writer) < 0)
          throw "closing <trkpt> element";
      }
      if (xmlTextWriterEndElement(writer) < 0)
          throw "closing <trkpt> element";

      if (xmlTextWriterEndElement(writer) < 0)
          throw "closing <trk> element";
    }

    if (xmlTextWriterEndDocument(writer) < 0)
      throw "closing xml document";
  }
  catch (const char *c){
    xmlFreeTextWriter(writer);
    throw Err() << "write_gpx: error in " << c;
  }

  // free resources
  xmlFreeTextWriter(writer);

  return;
}

/********************************************************************/
#define TYPE_ELEM      1
#define TYPE_TEXT      3
#define TYPE_COMM      8
#define TYPE_ELEM_END 15
#define TYPE_SWS      14

#define NAMECMP(x) (xmlStrcasecmp(name,(const xmlChar *)x)==0)
#define GETATTR(x) (const char *)xmlTextReaderGetAttribute(reader, (const xmlChar *)x)
#define GETVAL     (const char *)xmlTextReaderConstValue(reader)

// read <extensions>. In gpx/track/rte/wpt/trkseg
int
read_ext_node(xmlTextReaderPtr reader, Opt & opts){
  string state = "";
  while(1){
    int ret =xmlTextReaderRead(reader);
    if (ret != 1) return ret;

    const xmlChar *name = xmlTextReaderConstName(reader);
    int type = xmlTextReaderNodeType(reader);

    if (type == TYPE_SWS || type == TYPE_COMM) continue;

    else if (type == TYPE_ELEM_END) {
      if (NAMECMP("extensions")) break;
    }

  }
  return 1;
}


// read waypoint / rte point
int
read_wpt_node(xmlTextReaderPtr reader, GeoWptList & data){
  GeoWpt wpt;
  auto a_x = GETATTR("lon");
  auto a_y = GETATTR("lat");
  wpt.x = a_x? atof(a_x):0;
  wpt.y = a_y? atof(a_y):0;
  string state="";

  if (xmlTextReaderIsEmptyElement(reader)) {
    data.push_back(wpt);
    return 1;
  }

  while(1){
    int ret =xmlTextReaderRead(reader);
    if (ret != 1) return ret;

    const xmlChar *name = xmlTextReaderConstName(reader);
    int type = xmlTextReaderNodeType(reader);

    if (type == TYPE_SWS || type == TYPE_COMM) continue;

    else if (NAMECMP("extensions") && (type == TYPE_ELEM)){
      cerr << "Warning: skip <extensions> in <wpt>/<rtept>\n";
      ret = read_ext_node(reader, data.opts);
      if (ret != 1) return ret;
    }

    else if (type == TYPE_ELEM) {
      state = "";
      if (NAMECMP("ele"))  state = "ele";
      if (NAMECMP("time")) state = "time";
      if (NAMECMP("name")) state = "name";
      if (NAMECMP("cmt"))  state = "comm";
      for (const char **fn = gps_wpt_names; *fn!=NULL; fn++){
        if (NAMECMP(*fn)) state = *fn;
      }
      if (state == "")
         cerr << "Warning: Unknown node \"" << name << "\" in wpt (type: " << type << ")\n";
    }
    else if (type == TYPE_ELEM_END) {
      if (NAMECMP("ele") || NAMECMP("time") || NAMECMP("name") || NAMECMP("cmt")) state = "";
      for (const char **fn = gps_wpt_names; *fn!=NULL; fn++){
        if (NAMECMP(*fn)) state = "";
      }
      if (NAMECMP("wpt") || NAMECMP("rtept")) break;
      if (state != "")
         cerr << "Warning: Unknown node \"" << name << "\" in wpt (type: " << type << ")\n";
    }

    else if (type == TYPE_TEXT){
      if (state == "ele")  wpt.z = atof(GETVAL);
      if (state == "time") wpt.t = parse_utc_time(GETVAL);
      if (state == "name") wpt.name = GETVAL;
      if (state == "comm") wpt.comm = GETVAL;
      for (const char **fn = gps_wpt_names; *fn!=NULL; fn++){
        string pfn = string("gpx_") + (*fn);
        if (state == *fn) wpt.opts.put<string>(pfn, GETVAL);
      }
    }
    else {
      cerr << "Warning: Unknown node \"" << name << "\" in wpt (type: " << type << ")\n";
    }
  }
  data.push_back(wpt);
  return 1;
}

int
read_trkpt_node(xmlTextReaderPtr reader, GeoTrk & trk, bool start){
  GeoTpt pt;
  bool is_ele = false, is_time=false;

  auto a_x = GETATTR("lon");
  auto a_y = GETATTR("lat");
  pt.x = a_x? atof(a_x):0;
  pt.y = a_y? atof(a_y):0;

  if (xmlTextReaderIsEmptyElement(reader)) {
    pt.start = start;
    trk.push_back(pt);
    return 1;
  }

  while(1){
    int ret =xmlTextReaderRead(reader);
    if (ret != 1) return ret;

    const xmlChar *name = xmlTextReaderConstName(reader);
    int type = xmlTextReaderNodeType(reader);

    if (type == TYPE_SWS || type == TYPE_COMM) continue;

    else if (NAMECMP("extensions") && (type == TYPE_ELEM)){
      Opt o;
      cerr << "Warning: skip <extensions> in <trkpt>\n";
      ret = read_ext_node(reader, o);
      if (ret != 1) return ret;
    }

    else if (NAMECMP("ele")){
      if (type == TYPE_ELEM) is_ele = true;
      if (type == TYPE_ELEM_END) is_ele = false;
    }
    else if (NAMECMP("time")){
      if (type == TYPE_ELEM) is_time = true;
      if (type == TYPE_ELEM_END) is_time = false;
    }
    else if (type == TYPE_TEXT){
      if (is_ele)
        pt.z = atof(GETVAL);
      if (is_time)
        pt.t = parse_utc_time(GETVAL);
    }

    else if (NAMECMP("trkpt") && type == TYPE_ELEM_END){
      break;
    }
    else {
      cerr << "Warning: Unknown node \"" << name << "\" in trkpt (type: " << type << ")\n";
    }
  }
  pt.start = start;
  trk.push_back(pt);
  return 1;
}


int
read_trkseg_node(xmlTextReaderPtr reader, GeoTrk & trk){
  bool start=true;
  while(1){
    int ret =xmlTextReaderRead(reader);
    if (ret != 1) return ret;

    const xmlChar *name = xmlTextReaderConstName(reader);
    int type = xmlTextReaderNodeType(reader);

    if (type == TYPE_SWS || type == TYPE_COMM) continue;

    else if (NAMECMP("extensions") && (type == TYPE_ELEM)){
      Opt o;
      cerr << "Warning: skip <extensions> in <trkseg>\n";
      ret = read_ext_node(reader, o);
      if (ret != 1) return ret;
    }

    else if (NAMECMP("trkpt") && (type == TYPE_ELEM)){
      ret = read_trkpt_node(reader, trk, start);
      if (ret != 1) return ret;
      start=false;
    }

    else if (NAMECMP("trkseg") && (type == TYPE_ELEM_END)){
      break;
    }

    else {
      cerr << "Warning: Unknown node \"" << name << "\" in trkseg (type: " << type << ")\n";
    }
  }
  return 1;
}

int
read_trk_node(xmlTextReaderPtr reader, GeoData & data, const Opt & opts){
  GeoTrk trk;
  string state;
  while(1){
    int ret =xmlTextReaderRead(reader);
    if (ret != 1) return ret;

    const xmlChar *name = xmlTextReaderConstName(reader);
    int type = xmlTextReaderNodeType(reader);

    if (type == TYPE_SWS || type == TYPE_COMM) continue;

    else if (NAMECMP("extensions") && (type == TYPE_ELEM)){
      cerr << "Warning: skip <extensions> in <trk>\n";
      ret = read_ext_node(reader, trk.opts);
      if (ret != 1) return ret;
    }

    else if (NAMECMP("trkseg") && (type == TYPE_ELEM)){
      ret=read_trkseg_node(reader, trk);
      if (ret != 1) return ret;
    }

    else if (type == TYPE_ELEM) {
      state = "";
      if (NAMECMP("cmt"))  state = "comm";
      if (NAMECMP("name"))  state = "name";
      for (const char **fn = gps_trk_names; *fn!=NULL; fn++){
        if (NAMECMP(*fn)) state = *fn;
      }
      if (state == "")
         cerr << "Warning: Unknown node \"" << name << "\" in trk (type: " << type << ")\n";
    }

    else if (type == TYPE_ELEM_END) {
      if (NAMECMP("cmt") || NAMECMP("name")) state = "";
      for (const char **fn = gps_trk_names; *fn!=NULL; fn++){
        if (NAMECMP(*fn)) state = "";
      }
      if (NAMECMP("trk")) break;
      if (state != "")
         cerr << "Warning: Unknown node \"" << name << "\" in trk (type: " << type << ")\n";
    }

    else if (type == TYPE_TEXT){
      if (state == "name") trk.name = GETVAL;
      if (state == "comm") trk.comm = GETVAL;
      for (const char **fn = gps_trk_names; *fn!=NULL; fn++){
        if (state == *fn) trk.opts.put<string>(*fn, GETVAL);
      }
    }

    else {
      cerr << "Warning: Unknown node \"" << name << "\" in trk (type: " << type << ")\n";
    }
  }
  data.trks.push_back(trk);
  if (opts.get("verbose", false))
    cerr << "  Reading track: " << trk.name
         << " (" << trk.size() << " points)" << endl;
  return 1;
}

int
read_rte_node(xmlTextReaderPtr reader, GeoData & data, const Opt & opts){
  GeoWptList wptl;
  string state;
  while(1){
    int ret =xmlTextReaderRead(reader);
    if (ret != 1) return ret;

    const xmlChar *name = xmlTextReaderConstName(reader);
    int type = xmlTextReaderNodeType(reader);

    if (type == TYPE_SWS || type == TYPE_COMM) continue;

    else if (NAMECMP("extensions") && (type == TYPE_ELEM)){
      cerr << "Warning: skip <extensions> in <rte>\n";
      ret = read_ext_node(reader, wptl.opts);
      if (ret != 1) return ret;
    }

    // points are same as wpt
    else if (NAMECMP("rtept") && (type == TYPE_ELEM)){
      ret=read_wpt_node(reader, wptl);
      if (ret != 1) return ret;
    }

    // fields are same as in trk!
    else if (type == TYPE_ELEM) {
      state = "";
      if (NAMECMP("name")) state = "name";
      if (NAMECMP("cmt"))  state = "comm";
      for (const char **fn = gps_trk_names; *fn!=NULL; fn++){
        if (NAMECMP(*fn)) state = *fn;
      }
      if (state == "")
         cerr << "Warning: Unknown node \"" << name << "\" in rte (type: " << type << ")\n";
    }

    else if (type == TYPE_ELEM_END) {
      if (NAMECMP("name")) state = "";
      if (NAMECMP("cmt"))  state = "";
      for (const char **fn = gps_trk_names; *fn!=NULL; fn++){
        if (NAMECMP(*fn)) state = "";
      }
      if (NAMECMP("rte")) break;
      if (state != "")
         cerr << "Warning: Unknown node \"" << name << "\" in rte (type: " << type << ")\n";
    }

    else if (type == TYPE_TEXT){
      if (state == "name") wptl.name = GETVAL;
      if (state == "comm") wptl.comm = GETVAL;
      for (const char **fn = gps_trk_names; *fn!=NULL; fn++){
        if (state == *fn) wptl.opts.put<string>(*fn, GETVAL);
      }
    }

    else {
      cerr << "Warning: Unknown node \"" << name << "\" in rte (type: " << type << ")\n";
    }
  }
  data.wpts.push_back(wptl);
  if (opts.get("verbose", false))
    cerr << "  Reading route: " << wptl.name
         << " (" << wptl.size() << " points)" << endl;
  return 1;
}



int
read_gpx_node(xmlTextReaderPtr reader, GeoData & data,
              const Opt & opts, const string &fname){

  GeoWptList wptl;

  bool is_meta=false;
  while(1){
    int ret =xmlTextReaderRead(reader);
    if (ret != 1) return ret;

    const xmlChar *name = xmlTextReaderConstName(reader);
    int type = xmlTextReaderNodeType(reader);

    if (type == TYPE_SWS || type == TYPE_COMM) continue;

    // skip metadata
    else if (NAMECMP("metadata")){
      if (type == TYPE_ELEM) is_meta=true;
      if (type == TYPE_ELEM_END) is_meta=false;
    }
    else if (is_meta) continue;

    // read wpt,trk,rte tags
    else if (NAMECMP("wpt") && (type == TYPE_ELEM)){
      ret=read_wpt_node(reader, wptl);
      if (ret != 1) return ret;
    }
    else if (NAMECMP("trk") && (type == TYPE_ELEM)){
      ret=read_trk_node(reader, data, opts);
      if (ret != 1) return ret;
    }
    else if (NAMECMP("rte") && (type == TYPE_ELEM)){
      ret=read_rte_node(reader, data, opts);
      if (ret != 1) return ret;
    }

    // skip extensions
    else if (NAMECMP("extensions") && (type == TYPE_ELEM)){
      cerr << "Warning: skip <extensions> in <gpx>\n";
      ret=read_ext_node(reader, data.opts);
      if (ret != 1) return ret;
    }
    else if (NAMECMP("gpx") && (type == TYPE_ELEM_END)){
      break;
    }
    else {
      cerr << "Warning: Unknown node \"" << name << "\" in gpx (type: " << type << ")\n";
    }
  }
  if (wptl.size()){
    wptl.name = file_get_basename(fname, ".gpx");
    if (opts.get("verbose", false))
      cerr << "  Reading waypoints: " << wptl.name
           << " (" << wptl.size() << " points)" << endl;
    data.wpts.push_back(wptl);
  }
  return 1;
}


void
read_gpx(const string &filename, GeoData & data, const Opt & opts) {

  LIBXML_TEST_VERSION

  xmlTextReaderPtr reader;
  int ret;

  reader = xmlReaderForFile(filename.c_str(), NULL, 0);
  if (reader == NULL)
    throw Err() << "Can't open GPX file: " << filename;

  if (opts.get("verbose", false)) cerr <<
    "Reading GPX file: " << filename << endl;

  // parse file
  while (1){
    ret = xmlTextReaderRead(reader);
    if (ret!=1) break;

    const xmlChar *name = xmlTextReaderConstName(reader);
    int type = xmlTextReaderNodeType(reader);
    if (NAMECMP("gpx") && (type == TYPE_ELEM))
      ret = read_gpx_node(reader, data, opts, filename);
    if (ret!=1) break;
  }

  // free resources
  xmlFreeTextReader(reader);

  if (ret != 0) throw Err() << "Can't parse GPX file: " << filename;

}

